from BroachRpc.network import MsgHandler
from BroachRpc.common import GlobalVariable
from BroachRpc.exception import DefException
import socket
import logging
import time
import threading
from BroachRpc.common import CommonDef

@CommonDef.singleton
class RpcClient:

    def __init__(self):
        self.pool = GlobalVariable.BroachPool
        self.IntervalUdp = GlobalVariable.IntervalUdp # 重传间隔
        self.__NCPRev = {}  # 响应NCP消息
        self.__NCPRevLock = threading.RLock()

    @staticmethod
    def __send(data, ip, port):
        addr = (ip, int(port))
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        if len(data) > GlobalVariable.BufferLen :
            raise IOError("发送消息超过缓冲区3028字节长度限制")
        s.sendto(data, addr)
        #logging.debug("发送消息 -> "+str(addr)+" 内容: "+data.decode("utf-8"))
        s.close()

    def sendCIM(self, msg, ip, port):
        data = MsgHandler.getCIM(msg)
        self.__send(data, ip, int(port))

    def sendNCP(self, msgIdInfo, isS, ip, port):
        data = MsgHandler.getNCP(msgIdInfo, isS)
        self.__send(data, ip, port)

    def sendNES(self, msgId, msg, ip, port):
        if self.__NCPRev.get(msgId) is not None:
            return "idRepeat"
        msgDict, msgInfoList = MsgHandler.getNES(msgId, GlobalVariable.params["rpcIp"],
                                                 GlobalVariable.params["rpcPort"], msg)

        udpTimeOut = int(GlobalVariable.params["udpTimeOut"])
        timeUse = 0
        lock = threading.Condition()
        with self.__NCPRevLock:
            self.__NCPRev[msgId] = {"lock": lock, "msgInfo": None}

        while timeUse <= udpTimeOut:
            #time.sleep(GlobalVariable.IntervalUdp)
            lock.acquire()
            if timeUse == 0:
                for key in msgInfoList:
                    self.__send(msgDict[key], ip, port)
            lock.wait(GlobalVariable.IntervalUdp)
            lock.release()
            timeUse = timeUse + GlobalVariable.IntervalUdp
            revInfoList = self.__NCPRev.get(msgId).get("msgInfo")
            if revInfoList is not None:
                for revInfo in revInfoList:
                    msgInfo = revInfo[0]
                    isS = revInfo[1]
                    if isS == "0" and msgInfo in msgInfoList:
                        msgInfoList.remove(msgInfo)
                    elif isS == "2":#id重复 报错 重发
                        with self.__NCPRevLock:
                            self.__NCPRev.pop(msgId)
                        return "idRepeat"
            if len(msgInfoList) == 0:
                break

        with self.__NCPRevLock:
            self.__NCPRev.pop(msgId)
        if timeUse > udpTimeOut and len(msgInfoList) != 0:
            raise DefException.RpcSendNESTimeOutError("发送响应超时 ->"+ip+":"+str(port)+" 无应答")

    def handlerNCP(self, msgId, msgInfo, isS):
        with self.__NCPRevLock:
            if self.__NCPRev.get(msgId) is None:
                return
            lock = self.__NCPRev.get(msgId).get("lock")
            if self.__NCPRev.get(msgId).get("msgInfo") is None:
                msgInfoList = [[msgInfo, isS]]
                self.__NCPRev.get(msgId)["msgInfo"] = msgInfoList
            else:
                self.__NCPRev.get(msgId)["msgInfo"].append([msgInfo, isS])
        lock.acquire()
        lock.notify()
        lock.release()


from BroachRpc.threadPool import BroachPool
from BroachRpc.common import GlobalVariable
import json
import logging
from BroachRpc.network import RpcService

"""
框架主启动入口
"""
def run(path='./application.json'):
    # 初始化配置
    getConfig(path)
    # 初始化日志配置
    LOG_FORMAT = '%(asctime)s -%(name)s- %(threadName)s-%(thread)d - %(levelname)s - %(message)s'
    DATE_FORMAT = "%Y/%m/%d %H:%M:%S %p"
    logLevel = GlobalVariable.logLevel
    if logLevel == "info":
        logging.basicConfig(level=logging.INFO, format=LOG_FORMAT, datefmt=DATE_FORMAT)
    elif logLevel == "debug":
        logging.basicConfig(level=logging.DEBUG, format=LOG_FORMAT, datefmt=DATE_FORMAT)
    elif logLevel == "error":
        logging.basicConfig(level=logging.ERROR, format=LOG_FORMAT, datefmt=DATE_FORMAT)
    else:
        logging.basicConfig(level=logging.INFO, format=LOG_FORMAT, datefmt=DATE_FORMAT)
    logging.info("获取配置 " + json.dumps(GlobalVariable.params))
    # 初始化线程池
    threadParams = GlobalVariable.params["threadPool"]
    GlobalVariable.BroachPool = BroachPool.Pool(core=int(threadParams["core"]),
                                                threadMax=int(threadParams["threadMax"]),
                                                keepAliveTime=int(threadParams["keepAliveTime"]),
                                                queue=BroachPool.ArrayQueue(
                                                    queueMax=int(threadParams["queueMax"]),
                                                    isExcInfo=threadParams["isExcInfo"] == "True" or threadParams["isExcInfo"] == "true",
                                                    createThreadThreshold=int(threadParams["createThreadThreshold"])),
                                                poolName="BroachPool")
    RpcService.instance = RpcService.RpcService()
    # 启动服务端进程
    RpcService.instance.serverStart()

def getConfig(path):
    with open(path, encoding="utf-8") as f:
        getJson = json.load(f)
    if getJson.get("name") is not None:
        GlobalVariable.name = getJson.get("name")
    if getJson.get("version") is not None:
        GlobalVariable.version = getJson.get("version")
    if getJson.get("logLevel") is not None:
        GlobalVariable.logLevel = getJson.get("logLevel")
    if getJson.get("params") is not None:
        params = getJson.get("params")
        if params.get("rpcPort") is not None:
            GlobalVariable.params["rpcPort"] = params.get("rpcPort")
        if params.get("rpcIp") is not None:
            GlobalVariable.params["rpcIp"] = params.get("rpcIp")
        if params.get("clusterAddress") is not None:
            GlobalVariable.params["clusterAddress"] = params.get("clusterAddress")
        if params.get("rpcTimeOut") is not None:
            GlobalVariable.params["rpcTimeOut"] = params.get("rpcTimeOut")
        if params.get("udpTimeOut") is not None:
            GlobalVariable.params["udpTimeOut"] = params.get("udpTimeOut")
        if params.get("threadPool") is not None:
            threadPool = params.get("threadPool")
            if threadPool.get("core") is not None:
                GlobalVariable.params["threadPool"]["core"] = threadPool.get("core")
            if threadPool.get("threadMax") is not None:
                GlobalVariable.params["threadPool"]["threadMax"] = threadPool.get("threadMax")
            if threadPool.get("keepAliveTime") is not None:
                GlobalVariable.params["threadPool"]["keepAliveTime"] = threadPool.get("keepAliveTime")
            if threadPool.get("queueMax") is not None:
                GlobalVariable.params["threadPool"]["queueMax"] = threadPool.get("queueMax")
            if threadPool.get("isExcInfo") is not None:
                GlobalVariable.params["threadPool"]["isExcInfo"] = threadPool.get("isExcInfo")
            if threadPool.get("createThreadThreshold") is not None:
                GlobalVariable.params["threadPool"]["createThreadThreshold"] = threadPool.get("createThreadThreshold")



import asyncio
import time
import threading

lock = threading.Condition()

async def work1(x):
    #lock.acquire()
    #lock.wait(2)
    #lock.release()
    print("work1-1: "+x)
    await asyncio.sleep(2)
    print("work1-2: "+x)
    return x

async def work2(x):
    print("work2-1: " + x)

    print("work2-2: " + x)
    return x

def callback(future):
    print("获得返回值: "+ future.result())

a = work1("1")
b = work2("2")
loop = asyncio.get_event_loop() # 创建事件循环
#task = asyncio.wait([a,b])# 创建task任务
task1 = loop.create_task(a)
task1.add_done_callback(callback)
task2 = loop.create_task(b)
task2.add_done_callback(callback)
tasks = asyncio.wait([task1, task2])
loop.run_until_complete(tasks) # 把协程对象丢给循环,并执行异步函数内部代码
asyncio.Queue()


import random
import threading
import time
import logging
from BroachRpc.common import GlobalVariable

"""
单例装饰器
"""
def singleton(cls):
    """
    单例装饰器
    :param cls: 类
    :return: 实例
    """
    _instance_dict = {}  # 采用字典，可以装饰多个类，控制多个类实现单例模式
    lock = threading.RLock()
    def inner(*args, **kwargs):
        if cls in _instance_dict:
            return _instance_dict.get(cls)
        with lock:
            if cls not in _instance_dict:
                _instance_dict[cls] = cls(*args, **kwargs)
            return _instance_dict.get(cls)
    return inner

"""
id获取
"""
def getId(n=9):
    """
    生成随机id 默认9位
    :param n: 位数
    :return: 字符串id
    """
    id = ""
    template = "0987654321qwertyuioplkjhgfdsazxcvbnmQWERTYUIOPASDFGHJKLZXCVBNM"
    for i in range(0,n):
        strIndex = random.randint(0,61)
        id = id + template[strIndex]
    return id

"""
定时任务执行函数
"""
def timerTask(tasks:list):
    while True:
        time.sleep(1)
        for item in tasks:
            fn = item.get("fn")
            timeStep = int(item.get("timeStep"))
            lastTime = item.get("time")
            if lastTime is None:
                item["time"] = time.time()
                fn()
            else:
                ex = time.time() - lastTime
                if ex >= timeStep:
                    item["time"] = time.time()
                    fn()


from BroachRpc.common import GlobalVariable, CommonDef
from BroachRpc.network import MsgHandler
from BroachRpc.threadPool import BroachPool
import socket
import logging
import threading

@CommonDef.singleton
class RpcServer:
    def __init__(self, rpcClient):
        self.address=(GlobalVariable.params["rpcIp"], GlobalVariable.params["rpcPort"])
        self.recvLen = GlobalVariable.BufferLen
        self.threadPool = BroachPool.Pool(core=5,
                                          threadMax=10,
                                          keepAliveTime=10,
                                          queue=BroachPool.ArrayQueue(
                                              queueMax=10000,
                                              isExcInfo=False,
                                              createThreadThreshold=5),
                                          poolName="rpcServerPool")
        self.broachPool = GlobalVariable.BroachPool
        self.__callBackNES = []
        self.__callBackCIM = []
        self.__NESRevPart = {} #收集收到的NES消息片段
        self.__NESRevLock = threading.RLock()
        self.__rpcClient = rpcClient
    def start(self):
        self.threadPool.submit(self.__serverRun)

    def addNESCallBack(self, fn):
        self.__callBackNES.append(fn)

    def addCIMCallBack(self, fn):
        self.__callBackCIM.append(fn)

    def __serverRun(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.bind(self.address)
        logging.info("rpc 服务端口启动 -> "+str(self.address))
        while True:
            try:
                data, addr = s.recvfrom(self.recvLen)
                self.threadPool.submit(self.__msgHandler, data, addr)
            except :
                logging.exception(exc_info=True, msg="rpc 通信异常")

    def __msgHandler(self, data, addr):
        #logging.debug("获得消息 <- " + str(addr)+" 内容: "+data.decode("utf-8"))
        msgType, msg = MsgHandler.decodeData(data)
        if msgType == "NCP":
            self.__NCPHandler(msg)
        elif msgType == "NES":
            self.__NESHandler(msg)
        elif msgType == "CIM":
            self.__CIMHandler(msg)

    def __CIMHandler(self, msg):
        for fn in self.__callBackCIM:
            fn(MsgHandler.decodeCIM(msg))

    def __NCPHandler(self, msg):
        msgId, msgInfo, isS = MsgHandler.decodeNCP(msg)
        self.__rpcClient.handlerNCP(msgId, msgInfo, isS)

    def __NESHandler(self, msg):
        msgId, msgInfo, revIp, revPort, msgPart = MsgHandler.decodeNES(msg)
        msgOrder = int(msgInfo[5:10])
        msgLen = int(msgInfo[0:5])
        if msgLen == 1:
            #无分段消息
            self.__rpcClient.sendNCP(msgId+msgInfo, "0", revIp, revPort)
            for fn in self.__callBackNES:
                self.broachPool.submit(fn, msgPart.decode("utf-8"), revIp, int(revPort))
            return
        with self.__NESRevLock:
            msgPartDict = self.__NESRevPart.get(msgId)
            if msgPartDict is None:
                self.__NESRevPart[msgId] = {msgOrder:msgPart, "msgLen": msgLen}
                self.__rpcClient.sendNCP(msgId+msgInfo, "0", revIp, revPort)
            else:
                if msgPartDict.get(msgOrder) is None:
                    msgPartDict[msgOrder] = msgPart
                    self.__rpcClient.sendNCP(msgId+msgInfo, "0", revIp, revPort)
                else:
                    if msgPartDict[msgOrder] == msgPart:
                        self.__rpcClient.Client.sendNCP(msgId+msgInfo, "0", revIp, revPort)
                    else:
                        #RpcHandler.NESRevPart.pop(msgId)
                        self.__rpcClient.sendNCP(msgId+msgInfo, "2", revIp, revPort)
            msgPartDict = self.__NESRevPart.get(msgId)
            if (len(msgPartDict)-1) == msgLen:
                revMsg = b""
                for i in range(0, msgLen):
                    revMsg = revMsg + msgPartDict[i]
                #RpcHandler.NESRev[msgId] = {"msg": revMsg, "time":time.time()}
                #通知上游函数处理
                for fn in self.__callBackNES:
                    self.broachPool.submit(fn, revMsg.decode("utf-8"), revIp, int(revPort))
                #    fn(revMsg.decode("utf-8"))
                #删除片段中的NES消息
                self.__NESRevPart.pop(msgId)




import threading
from BroachRpc.exception.DefException import ReceiveBuffOverError
from BroachRpc.common import GlobalVariable
import time
"""
定义消息 
 消息类型3位 
 NES 消息 - 双向应答
 消息id  报文Id（18位） + 消息序号 00000 + 00000 10位 前五位表示总消息数量 后五位表示当前消息序号
 接收响应消息地址 15位ip地址 5位端口号 127.  0.  0.  118080 -> 127.0.0.1 18080
 固定字节共 3 + 18 + 5 + 5 + 15 + 5 = 51 位
 报文长度 = 3028 - 51 = 2977
 NCP 消息 - 响应NES消息
 消息id  发送方id 28位
 处理结果 1位 0 成功 1 失败 2 消息id当前已存在
 CIM 消息 
 缓冲区字节以内的单向udp消息
"""

def decodeData(data):
    msgType = data[0:3]
    msg = data[3:len(data)]
    return msgType.decode("utf-8"), msg

def getCIM(msg):
    data = "CIM" + msg
    data = data.encode("utf-8")
    if len(data) > GlobalVariable.BufferLen:
        raise ReceiveBuffOverError("CIM 消息不能超过 3025字节")
    return data

def decodeCIM(data):
    return data.decode("utf-8")

def getNCP(msgIdInfo, isS="0"):
    data = "NCP" + msgIdInfo + isS
    data = data.encode("utf-8")
    return data

def decodeNCP(data):
    msgId = data[0:18]
    msgInfo = data[18:28]
    isS = data[28:29]
    return msgId.decode("utf-8"), msgInfo.decode("utf-8") ,isS.decode("utf-8")

def getNES(msgId, revIp, revPort ,msg):
    msgDict = {}
    msgInfoList = []
    msgData = msg.encode("utf-8")
    msgLenLimit = GlobalVariable.BufferLen-51
    msgLen = len(msgData)
    if msgLen > msgLenLimit:
        y = msgLen%msgLenLimit
        if y == 0:
            order = int(msgLen/msgLenLimit)
        else:
            order = int(msgLen/msgLenLimit) + 1
        revAddress = getIpPortStr(revIp, revPort)
        orderS = getCover(str(order), 5, "0")
        for i in range(0, order):
            data = "NES" + msgId + orderS + getCover(str(i), 5, "0") + revAddress
            msgInfoList.append(orderS + getCover(str(i), 5, "0"))
            data = data.encode("utf-8")
            start = i*msgLenLimit
            end = (i+1) * msgLenLimit
            if end > msgLen:
                end = msgLen
            msgDict[orderS + getCover(str(i), 5, "0")] = data+msgData[start:end]
    else:
        data = "NES" + msgId + "00001" + "00000" + getIpPortStr(revIp, revPort)
        data = data.encode("utf-8")
        msgDict["00001" + "00000"] = data+msgData
        msgInfoList.append("00001" + "00000")
    return msgDict, msgInfoList

def decodeNES(data):
    msgId = data[0:18]
    msgInfo = data[18:28]
    revIp = data[28:43].decode("utf-8")
    revPort = data[43:48].decode("utf-8")
    msgPart = data[48:]
    return msgId.decode("utf-8"), msgInfo.decode("utf-8"), revIp.replace(" ",""), int(revPort), msgPart

def getIpPortStr(ip:str, port):
    ipl = ip.split(".")
    ips = ""
    for item in ipl:
        ips = ips + getCover(item, 3, " ") + "."
    ips = ips[0:15]
    return ips + getCover(str(port), 5, "0")

def getCover(s, l:int, coverStr):
    ls = len(s)
    lx = l - ls
    if lx <= 0:
        return s
    else:
        re = s
        for i in range(0, lx):
            re = coverStr+re
        return re
